import { HttpException, HttpStatus, Injectable } from "@nestjs/common";
import { Constants } from "src/utils/constants";
import { CreateProductServiceDto } from "./dto/create-product-service.dto";
import { UpdateProductServiceDto } from "./dto/update-product-service.dto";
import { ProductServiceRepository } from "./product-service.repository";
import { ResponseUtils } from "src/utils/response.utils";
import { ProductService } from "./entities/product-service.entity";
import { getGraphQLClientWithApiKey } from "src/helper/graphql-client-helper";
import {
    fetchAllProductsForScheduler,
    fetchAllRetailers,
    fetchOneProduct,
    fetchProductsCount
} from "src/helper/dutchie-plus-query";
import { ElasticsearchService } from "@nestjs/elasticsearch";
import { SearchProductServiceDto } from "./dto/search-product-service.dto";
import { InjectRepository } from "@nestjs/typeorm";
import { ProductCategoryRepository } from "../product-category/product-category.repository";
import { urlSlug } from "src/helper/common-helper";
import { RetailerServiceRepository } from "../retailer-service/retailer-service.repository";
import { ProductEffectsRepository } from "../product-effects/product-effects.repository";

@Injectable()
export class ProductServiceService {
  constructor(
    private readonly repository: ProductServiceRepository,
    private readonly esService: ElasticsearchService,
    @InjectRepository(ProductCategoryRepository)
    private readonly productCatRepo: ProductCategoryRepository,
    @InjectRepository(RetailerServiceRepository)
    private readonly retailerServiceRepo: RetailerServiceRepository,
    @InjectRepository(ProductEffectsRepository)
    private readonly productEffectRepo: ProductEffectsRepository
  ) {
  }

  async processProductData(productData: any, retailerId: string, retailerName: string): Promise<any | Error> {
    try {
      const [
        productService,
        productCategory,
        productEffects
      ] = await Promise.all([
        this.repository.findByFilterQuery({}),
        this.productCatRepo.findAll(),
        this.productEffectRepo.findByFilterQuery({})
      ]);

      const productServiceDatasetIds = new Set(productService.map((item) => item.productId));
      const productMapService: any = new Map(productService.map(item => [item.productId, item]));
      const productDifferences = [];
      const differeneceDeleteIds = [];
      const categoriesArr = [];
      const effectsArr = [];

      await Promise.all(productData.menu.products.map(async (item) => {
        // product service
        if (productServiceDatasetIds.size > 0) {
          if (!productServiceDatasetIds.has(item.id)) {
            productDifferences.push(item);
          } else {
            // Now you can find a product by id in constant time:
            const matchingItem = productMapService.get(item.id);
            // const matchingItem = productService.find((item2) => item2.productId === item.id);
            const brand = item.brand !== null ? item.brand.name : "";
            const brandImage = item.brand !== null ? item.brand.imageUrl : "";
            if (
              matchingItem &&
              (matchingItem.brandName !== brand ||
                matchingItem.brandImage !== brandImage ||
                matchingItem.image !== item.image ||
                matchingItem.productName !== item.name ||
                matchingItem.productSlug !== item.slug ||
                matchingItem.category !== item.category ||
                matchingItem.subCategory !== item.subcategory ||
                matchingItem.strainType !== item.strainType)
            ) {
              differeneceDeleteIds.push(item.id);
              productDifferences.push(item);
            }
          }
        } else {
          productDifferences.push(item);
        }

        // all categories
        categoriesArr.push(item.category);
        // all effects
        effectsArr.push(...item.effects);
      }));


      // if any update occers delete
      if (differeneceDeleteIds.length > 0) {
        await this.repository.deleteRowsFromDatabase(differeneceDeleteIds);
      }

      // create entities
      console.log(retailerName, ">", " New Product = ", productDifferences.length);
      const createProductServiceDtos = productDifferences.map((item) => {
        const brand = item.brand !== null ? item.brand.name : "";
        const brandImage = item.brand !== null ? item.brand.imageUrl : "";
        const brandId = item.brand !== null ? item.brand.id : "";
        return {
          retailerId,
          retailerName,
          productId: item.id,
          productName: item.name,
          productSlug: item.slug,
          image: item.image,
          brandName: brand,
          brandImage: brandImage,
          brandId: brandId,
          category: item.category,
          subCategory: item.subcategory,
          strainType: item.strainType,
          variants: item.variants.length
        };
      });

      if (createProductServiceDtos.length > 0) {
        await this.repository.createEntities(createProductServiceDtos);
      }

      // check effects
      const effects = Array.from(new Set(effectsArr));

      if (Object.keys(productEffects).length > 0) {
        const difference = effects.filter((effect) =>
          !productEffects.some((e) => e.effects === effect)
        );
        console.log("Product Effects Difference = ", difference);
        // if difference found then create entities with difference product effects
        const createProductEffectsDtos = difference.map((item) => {
          return {
            effects: item
          };
        });

        if (createProductEffectsDtos.length > 0) {
          await this.productEffectRepo.createEntities(createProductEffectsDtos);
        }
      } else {
        // if product effects table is empty in our local database then create all the effect in local
        const createProductEffectsDtos = effects.map((item) => {
          return {
            effects: item
          };
        });

        if (createProductEffectsDtos.length > 0) {
          await this.productEffectRepo.createEntities(createProductEffectsDtos);
        }
      }

      // check category
      const categories = Array.from(new Set(categoriesArr));

      if (Object.keys(productCategory).length > 0) {
        const difference = categories.filter((category) =>
          !productCategory.some((c) => c.categoryName === category)
        );
        console.log("Product Categories Difference = ", difference);
        // if difference found then create entities with difference product category
        const createProductCategoryDtos = difference.map((item) => {
          return {
            categoryName: item,
            urlSlug: urlSlug(item)
          };
        });

        if (createProductCategoryDtos.length > 0) {
          await this.productCatRepo.createEntities(createProductCategoryDtos);
        }
      } else {
        // if product categories table is empty in our local database then create all the category in local
        const createProductCategoryDtos = categories.map((item) => {
          return {
            categoryName: item,
            urlSlug: urlSlug(item)
          };
        });

        if (createProductCategoryDtos.length > 0) {
          await this.productCatRepo.createEntities(createProductCategoryDtos);
        }
      }

    } catch (e) {
      console.log(e);
      return e;
    }
  }

  async fetchRetailers(apiKeyName: string) {
    const graphQLClient = getGraphQLClientWithApiKey(apiKeyName);
    const query = fetchAllRetailers();
    const data = await graphQLClient.request(query);
    return data.retailers.map(retailer => ({ ...retailer, apiKeyName }));
  }

  async createUpdateRetailerDifference(): Promise<any | Error> {
    try {
      console.log(">>>Run Create or Update Dutchie Plus Retailers Scheduler<<<");
      const promises = [
        this.fetchRetailers("DUTCHIE_PLUS_API_TOKEN_1"),
        this.fetchRetailers("DUTCHIE_PLUS_API_TOKEN_2")
      ];
      const [retailers1, retailers2] = await Promise.all(promises);
      const data = [...retailers1, ...retailers2];

      const retailerService = await this.retailerServiceRepo.findByFilterQuery({});

      if (Object.keys(retailerService).length > 0) {
        const existingRetailerIds = retailerService.map(x => x.retailerId);
        const difference = data.filter(obj => !existingRetailerIds.includes(obj.id));
        console.log("Retailer Difference =", difference);

        const dtoObjects = difference.map(obj => ({
          retailerId: obj.id,
          retailerName: obj.name,
          retailerAddress: JSON.stringify(obj.addressObject),
          retailerDeliverySettings: JSON.stringify(obj.deliverySettings),
          hours: JSON.stringify(obj.hours),
          apiKeyName: obj.apiKeyName
        }));

        await this.retailerServiceRepo.createEntities(dtoObjects);
      } else {
        const dtoObjects = data.map(obj => ({
          retailerId: obj.id,
          retailerName: obj.name,
          retailerAddress: JSON.stringify(obj.addressObject),
          retailerDeliverySettings: JSON.stringify(obj.deliverySettings),
          hours: JSON.stringify(obj.hours),
          apiKeyName: obj.apiKeyName
        }));

        await this.retailerServiceRepo.createEntities(dtoObjects);
      }
      console.log(">>>End Create or Update Dutchie Plus Retailers Scheduler<<<");

    } catch (e) {
      console.log(e);
      return e;
    }
  }


  async createUpdateDutchiePlusProductsAndCategory() {
    try {
      console.log(">>>Run Create or Update Dutchie Plus Products Scheduler<<<");
      const promises = [
        this.fetchRetailers("DUTCHIE_PLUS_API_TOKEN_1"),
        this.fetchRetailers("DUTCHIE_PLUS_API_TOKEN_2")
      ];
      const [retailers1, retailers2] = await Promise.all(promises);
      const mergedRetailers = [...retailers1, ...retailers2];

      // Check if retailers exist
      if (mergedRetailers.length > 0) {
        // Create an array of promises for fetching product data
        console.log("Start Fetching product count Parallel  =", new Date());
        const fetchProductsCountObject = fetchProductsCount();
        const retailerListPromises = mergedRetailers.map((retailer) => {
          const graphQLClient = getGraphQLClientWithApiKey(retailer.apiKeyName);
          return graphQLClient.request(fetchProductsCountObject, { retailerId: retailer.id })
            .then(response => {
              if (response && response.menu && response.menu.productsCount != null) {

                return {
                  retailerId: retailer.id,
                  retailerName: retailer.name,
                  graphQLClient: graphQLClient,
                  pagination: {
                    offset: 0,
                    limit: response.menu.productsCount
                  }
                };
              }
            });
        });

        let allRetailerProductCount = await Promise.all(retailerListPromises);
        allRetailerProductCount = allRetailerProductCount.filter((data: any) => data.pagination.limit > 0);
        console.log("End Fetching product count Parallel =", new Date());


        console.log("Start Fetching product Parallel End =", new Date());
        const fetchProductsQuery = fetchAllProductsForScheduler();
        const fetchProductPromises = allRetailerProductCount.map(variable =>
          variable.graphQLClient.request(fetchProductsQuery, variable).then(response => {
            if (response && response.menu && response.menu.products) {
              return {
                retailerId: variable.retailerId,
                retailerName: variable.retailerName,
                productData: response
              };
            }
          })
        );
        console.log("End Fetching product Parallel End =", new Date());

        const allProductData = await Promise.all(fetchProductPromises);
        //console.log(allProductData)
        for (const data of allProductData) {
          console.log("Send for Processing", new Date());
          await this.processProductData(data.productData, data.retailerId, data.retailerName);
          console.log("End processing =", new Date());
        }
        console.log(">>>End Create or Update Dutchie Plus Products Scheduler<<<");
      } else {
        console.log("Retailers Not Found!");
      }
    } catch (e) {
      console.log(e);
      return e;
    }
  }

  async productSynchronize(): Promise<any | Error> {
    try {
      await this.createUpdateDutchiePlusProductsAndCategory();
      return ResponseUtils.successResponseHandler(200, "Product Synchronize completed Successfully!");
    } catch (e) {
      console.log(e);
      return e;
    }
  }

  async create(dto: CreateProductServiceDto): Promise<any | Error> {
    const data = await this.repository.createEntity(dto);
    if (!data) {
      throw new HttpException(Constants.CREATE_FAILED, HttpStatus.BAD_REQUEST);
    }
    return ResponseUtils.successResponseHandler(
      201,
      "Products created successfully",
      "data",
      data
    );
  }

  async findAll(): Promise<ProductService[] | Error> {
    try {
      const data = await this.repository.findAll();
      if (!data) {
        throw new HttpException(Constants.NOT_FOUND, HttpStatus.NOT_FOUND);
      }
      return ResponseUtils.successResponseHandler(
        200,
        "Products list fetched successfully",
        "data",
        data
      );
    } catch (error) {
      console.log(error);
      return error;
    }
  }

  async getRetailerWiseBrand(retailerId: string): Promise<ProductService[] | Error> {
    try {
      const data = await this.repository.find({
        where: {
          retailerId: retailerId,
          isDeleted: false
        }
      });
      if (!data) {
        throw new HttpException(Constants.NOT_FOUND, HttpStatus.NOT_FOUND);
      }

      const uniqueBrands = data
        .map(({ id, retailerId, retailerName, brandId, brandName, brandImage, isShowBrand }) => ({
          id,
          retailerId,
          retailerName,
          brandId,
          brandName,
          brandImage,
          isShowBrand
        })) // Extract specific properties
        .filter((item, index, self) => item.brandName !== "" && self.findIndex((t) => t.brandName === item.brandName) === index); // Filter empty brand names and duplicates based on brand

      return ResponseUtils.successResponseHandler(
        200,
        "Retailer Wise Brand list fetched successfully",
        "data",
        uniqueBrands
      );
    } catch (error) {
      console.log(error);
      return error;
    }
  }

  async findOne(slug: string, retailerId: string): Promise<any | Error> {
    try {
      const data = await this.repository.findOneEntity(slug, retailerId);
      if (Object.keys(data).length > 0) {
        const variables = {
          retailerId: data.retailerId,
          productId: data.productId
        };
        const retailerService = await this.retailerServiceRepo.findOne({
          where: {
            retailerId: data.retailerId,
            isDeleted: false
          }
        });
        if (retailerService) {
          const graphQLClient = getGraphQLClientWithApiKey(retailerService.apiKeyName);
          const query = fetchOneProduct();
          return await graphQLClient.request(query, variables);
        } else {
          await this.createUpdateRetailerDifference();
          return [];
        }
      } else {
        console.log(">>Fetch Products From Dutchie<<");
        await this.createUpdateDutchiePlusProductsAndCategory();
        await this.findOne(slug, retailerId);
      }
    } catch (error) {
      console.log(error);
      return error;
    }
  }

  async update(
    id: string,
    dto: UpdateProductServiceDto
  ): Promise<ProductService | Error> {
    const data = await this.repository.updateProductCategory(id, dto);
    if (!data) {
      throw new HttpException(Constants.UPDATE_FAILED, HttpStatus.BAD_REQUEST);
    }
    return ResponseUtils.successResponseHandler(
      200,
      "Products updated successfully",
      "data",
      data
    );
  }

  async showHideBrand(
    id: string,
    dto: any
  ): Promise<any | Error> {
    const data = await this.repository.update(id, {
      isShowBrand: dto.isShowBrand
    });
    if (!data) {
      throw new HttpException(Constants.UPDATE_FAILED, HttpStatus.BAD_REQUEST);
    }
    return ResponseUtils.successResponseHandler(
      200,
      "Brand data updated successfully!",
      "data",
      data
    );
  }

  async remove(id: string): Promise<boolean | Error> {
    const data = await this.repository.deleteEntity(id);
    if (!data) {
      throw new HttpException(Constants.NOT_FOUND, HttpStatus.BAD_REQUEST);
    }
    return ResponseUtils.buildDeletedData(
      data,
      200,
      "Products deleted successfully"
    );
  }

  //#region Elastic Search

  public async indexIsExists(index: string) {
    return await this.esService.indices.exists({ index });
  }

  public async deleteIndex(index: string) {
    const isExists = await this.indexIsExists(index);
    if (isExists) {
      await this.esService.indices.delete({ index });
    }
  }

  public async createIndex(index: string, body: any) {
    const isExists = await this.indexIsExists(index);
    if (!isExists) {
      const data = await this.esService.indices.create({ index, body });
      return data.acknowledged;
    }
    return false;
  }

  public async indexData(index: string, body: any) {
    return await this.esService.index({ index, body });
  }

  private async fetchAllProductsForElasticSearch() {
    const products = [];
    try {
      console.log(">>>Run Fetch All Products For Elastic Search<<<");

      const promises = [
        this.fetchRetailers("DUTCHIE_PLUS_API_TOKEN_1"),
        this.fetchRetailers("DUTCHIE_PLUS_API_TOKEN_2")
      ];
      const [retailers1, retailers2] = await Promise.all(promises);
      const mergedRetailers = [...retailers1, ...retailers2];

      // Check if retailers exist
      if (mergedRetailers.length > 0) {
        // Create an array of promises for fetching product data
        console.log("Start Fetching product count Parallel  =", new Date());
        const fetchProductsCountObject = fetchProductsCount();
        const retailerListPromises = mergedRetailers.map((retailer) => {
          const graphQLClient = getGraphQLClientWithApiKey(retailer.apiKeyName);
          return graphQLClient.request(fetchProductsCountObject, { retailerId: retailer.id })
            .then(response => {
              if (response && response.menu && response.menu.productsCount != null) {
                return {
                  retailerId: retailer.id,
                  retailerName: retailer.name,
                  graphQLClient: graphQLClient,
                  pagination: {
                    offset: 0,
                    limit: response.menu.productsCount
                  }
                };
              }
            });
        });

        let allRetailerProductCount = await Promise.all(retailerListPromises);
        allRetailerProductCount = allRetailerProductCount.filter((data: any) => data.pagination.limit > 0);
        console.log("End Fetching product count Parallel =", new Date());

        console.log("Start Fetching product Parallel =", new Date());
        const fetchProductsQuery = fetchAllProductsForScheduler();
        const fetchProductPromises = allRetailerProductCount.map((variable: any) =>
          variable.graphQLClient.request(fetchProductsQuery, variable).then(response => {
            if (response && response.menu && response.menu.products) {
              return {
                retailerId: variable.retailerId,
                retailerName: variable.retailerName,
                productData: response
              };
            }
          })
        );
        console.log("End Fetching product Parallel =", new Date());

        const allProductData = await Promise.all(fetchProductPromises);
        //console.log(allProductData)
        for (const data of allProductData) {
          const product = data.productData.menu.products.map(x => {
            const { id, brand, name, slug, variants, ...restData } = x;
            const brandName = brand?.name ?? "";
            const brandNameAndProductNameAndCategoryName = name.trim().startsWith(brandName.trim())
              ? `${name} ${x.category}` : `${brandName} ${name} ${x.category}`;

            return {
              ...restData,
              productId: id,
              productName: name,
              productSlug: slug,
              brandName,
              variants: variants?.length ?? 0,
              retailerId: data.retailerId,
              retailerName: data.retailerName,
              brandNameAndProductNameAndCategoryName
            };
          });
          products.push(...product);
        }
        console.log(">>>End Fetch All Products For Elastic Search<<<");
        // console.table(products.length);
      } else {
        console.log("Retailers Not Found!");
      }
    } catch (e) {
      console.log(e);
      return e;
    } finally {
      return products;
    }
  }

  public async indexBulkData(indexValue: string) {
    // const res: any = await this.findAll();
    // const products: ProductService[] = res?.data;
    const products: Array<any> = await this.fetchAllProductsForElasticSearch();

    await this.deleteIndex(indexValue);

    const body = {
      settings: {
        analysis: {
          filter: {
            autocomplete_filter: {
              type: "edge_ngram",
              min_gram: 1,
              max_gram: 15
            }
          },
          analyzer: {
            autocomplete: {
              type: "custom",
              tokenizer: "standard",
              filter: ["lowercase", "autocomplete_filter"]
            }
          }
        }
      },

      mappings: {
        properties: {
          category: {
            type: "text",
            fields: {
              keyword: {
                type: "keyword",
                ignore_above: 256
              }
            }
          },
          subcategory: {
            type: "text",
            fields: {
              keyword: {
                type: "keyword",
                ignore_above: 256
              }
            }
          },
          strainType: {
            type: "text",
            fields: {
              keyword: {
                type: "keyword",
                ignore_above: 256
              }
            }
          },
          image: {
            type: "text",
            fields: {
              keyword: {
                type: "keyword",
                ignore_above: 512
              }
            }
          },
          productId: {
            type: "text",
            fields: {
              keyword: {
                type: "keyword",
                ignore_above: 256
              }
            }
          },
          productName: {
            type: "text",
            fields: {
              keyword: {
                type: "keyword",
                ignore_above: 256
              }
            }
          },
          brandName: {
            type: "text",
            fields: {
              keyword: {
                type: "keyword",
                ignore_above: 256
              }
            }
          },
          brandNameAndProductNameAndCategoryName: {
            type: "text",
            analyzer: "autocomplete",
            search_analyzer: "standard",
            fields: {
              keyword: {
                type: "keyword",
                ignore_above: 512
              }
            }
          },
          productSlug: {
            type: "text",
            fields: {
              keyword: {
                type: "keyword",
                ignore_above: 256
              }
            }
          },
          retailerId: {
            type: "text",
            fields: {
              keyword: {
                type: "keyword",
                ignore_above: 256
              }
            }
          },
          retailerName: {
            type: "text",
            fields: {
              keyword: {
                type: "keyword",
                ignore_above: 256
              }
            }
          },
          variants: {
            type: "long"
          }
        }
      }
    };
    console.log("Start Insert Data In Index =", new Date());
    await this.createIndex(indexValue, body);

    const operations = products.flatMap((doc) => [
      { index: { _index: indexValue } },
      doc
    ]);
    await this.esService.bulk({ refresh: true, operations });
    console.log("End Insert Data In Index =", new Date());

    return ResponseUtils.successResponseHandler(200, "Elastic DataBase Update Successfully.");
  }

  public async search(index: string, searchProductServiceDto: SearchProductServiceDto) {
    const { search, customField, page, limit, retailerId } = searchProductServiceDto;

    let exactsMatch = [];
    if (retailerId && retailerId != "") {
      const retailersMatch = {
        match_phrase: {
          retailerId: retailerId.toLowerCase()
        }
      };
      exactsMatch = [...exactsMatch, retailersMatch];
    }

    let fields: Array<string> = ["brandNameAndProductNameAndCategoryName"];
    if (customField && customField?.length) {
      fields = customField;
    }

    const result = await this.esService.search<any>({
      index,
      body: {
        query: {
          bool: {
            must: [
              ...exactsMatch,
              {
                match: {
                  brandNameAndProductNameAndCategoryName: {
                    query: search,
                    fuzziness: "AUTO",
                    fuzzy_transpositions: false,
                    prefix_length: 3
                  }
                }
              }
            ]
          }
        },
        highlight: {
          fields: {
            brandNameAndProductNameAndCategoryName: {
              pre_tags: ["<b>"],
              post_tags: ["</b>"]
            }
          }
        }
      },
      // from: page && page > 0 ? page - 1 : 0,
      size: limit && limit > 0 ? limit : 1000
    });
    const hits = result.hits.hits;
    const pageCount: any = result.hits.total;

    const products = hits.map((item: any) => {
      return {
        ...item._source,
        highlight: item?.highlight?.brandNameAndProductNameAndCategoryName?.[0]
      };
    });
    const data = { products, total: (pageCount?.value ?? pageCount) };

    if (!data) {
      throw new HttpException(Constants.NOT_FOUND, HttpStatus.NOT_FOUND);
    }
    return ResponseUtils.successResponseHandler(
      200,
      "Products list fetched successfully",
      "data",
      data
    );
  }

  public async searchDetails(index: string, searchProductServiceDto: SearchProductServiceDto) {
    const { search, customField, page, limit, retailerId } = searchProductServiceDto;

    let exactsMatch = [];
    if (retailerId && retailerId != "") {
      const retailersMatch = {
        match_phrase: {
          retailerId: retailerId.toLowerCase()
        }
      };
      exactsMatch = [...exactsMatch, retailersMatch];
    }

    let fields: Array<string> = ["brandNameAndProductNameAndCategoryName"];
    if (customField && customField?.length) {
      fields = customField;
    }

    const result = await this.esService.search<any>({
      index,
      body: {
        query: {
          bool: {
            must: [
              ...exactsMatch,
              {
                match: {
                  brandNameAndProductNameAndCategoryName: {
                    query: search,
                    fuzziness: "AUTO",
                    fuzzy_transpositions: false,
                    prefix_length: 3
                  }
                }
              }
            ]
          }
        }
      },
      // from: page ? page : 0,
      size: 2000
    });

    const data = result.hits.hits;
    const start = page ? page : 0;
    const end = limit > 0 ? limit : 20;
    const hits = data.slice(start * end, (start + 1) * end);
    const pageCount: any = result.hits.total;
    let menu = {
      /*
      weights: [],
      brands: [],
      subCategory: []
      */
      productsCount: pageCount?.value ?? pageCount,
      products: []
    };
    if (hits.length > 0) {
      const retailerService = await this.retailerServiceRepo.findOne({
        where: {
          retailerId: retailerId,
          isDeleted: false
        }
      });

      if (retailerService) {
        const graphQLClient = getGraphQLClientWithApiKey(retailerService.apiKeyName);
        /*
        let BrandHashSet = new Set();
        let subCategoryHashSet = new Set();
        let weightsHashSet = new Set();
        */
        const allApi = [];

        for (let i = 0; i < hits.length; i++) {
          const productId = hits[i]._source.productId;
          const variables = { retailerId, productId };
          const query = fetchOneProduct();
          const value = graphQLClient.request(query, variables);
          allApi.push(value.catch((error: any) => {
            console.error(error);
            return Promise.resolve();
          }));
        }
        // const results = await Promise.all(allApi);
        await Promise.all(allApi)
          .then((results) => {
            for (let j = 0; j < results.length; j++) {
              if (results?.[j]?.product) {
                menu.products.push(results?.[j]?.product);
              }
              /* <-- commented as it is not required in despo -->
              results[j].product.variants.forEach((x: { option: any; }) => {
                  if (!weightsHashSet.has(x.option)) {
                      weightsHashSet.add(x.option);
                      menu.weights.push(x.option);
                  }
              });
              if (!BrandHashSet.has(results[j].product.brand.id)) {
                  BrandHashSet.add(results[j].product.brand.id);
                  menu.brands.push(results[j].product.brand);
              }
              if (!subCategoryHashSet.has(results[j].product.subcategory)) {
                  subCategoryHashSet.add(results[j].product.subcategory);
                  menu.subCategory.push(results[j].product.subcategory);
              }
              */
            }
          })
          .catch(error => {
            console.log(error);
          })
          .finally(() => {
            return { menu };
          });
      } else {
        return { menu };
      }
    }

    return { menu };
  }

  //#endregion Elastic Search
}
